/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ServStats.
 *
 * The Initial Developer of the Original Code is
 * George J. Lee <gjl@csail.mit.edu>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{d9143379-0241-4a39-a424-63ab66b60b57}");
const CLASS_NAME = "CAPRI Diagnosis User Agent Component";
const CONTRACT_ID = "@capri.csail.mit.edu/ua;1";
const DATA_ID = "kpd1";

const STATE_START = Components.interfaces.nsIWebProgressListener.STATE_START;
const STATE_STOP  = Components.interfaces.nsIWebProgressListener.STATE_STOP;
const OK = 0;
const MAX_HIST_SIZE = 100;

const PROTO_VER = "0.2";
const CLIENT_VER = "0.12";

const DIAG_DIR = "diag";
const CONN_HIST_FILE = "conn-hist.xml";
const CONN_SUMMARY_FILE = "conn-summary.rdf";
const PREFIX = "http://kp.csail.mit.edu/capri";
const CONN_HIST_URI = PREFIX + "#connHist";
const CONN_SUMMARY_URI = PREFIX + "#connSummary";

const RDFC_ID = '@mozilla.org/rdf/container;1';
const RDFCUtils_ID = '@mozilla.org/rdf/container-utils;1';
var RDFCUtils = Components.classes[RDFCUtils_ID].getService(Components.interfaces.nsIRDFContainerUtils);

const RDF_ID = '@mozilla.org/rdf/rdf-service;1';
var RDF = Components.classes[RDF_ID].getService(Components.interfaces.nsIRDFService);
/*****************************************************************************
 * diag component
 *****************************************************************************/
function diagAgent() {};
diagAgent.prototype = {
  initialized: false,
  pref: null,
  obs: null,
  connHist: null,  // an associative array of arrays of connection entries
  connSummary: null,
  logging: false,
  histSize: 0,
  oldHistSize: 0,
  debug: false,
  
  // Keep track of the start time of each active connection to detect
  // unresponsive servers.
  startTime: null,

  dump: function(aMessage) {
    if (this.debug) {
      var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
      consoleService.logStringMessage("DiagAgent: " + aMessage);
    }
  },


  getContainer: function(ds, rootnode) {
    var con = Components.classes[RDFC_ID].createInstance(Components.interfaces.nsIRDFContainer);
    con.Init(ds, rootnode);
    return con;
  },

  getProfileFile: function(filename) {
    // get profile directory
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
    .getService(Components.interfaces.nsIProperties)
    .get("ProfD", Components.interfaces.nsIFile);

    file.append(DIAG_DIR);

    // if it doesn't exist, create
    if( !file.exists() ) {   
      file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0755);
    }

    file.append(filename);
    return file;
  },

  createDatasource: function(filename, uri, reset) {
    var ds;
    var file = this.getProfileFile(filename);
    var fileURI = this.getFileURI(file);

    // if it doesn't exist, create
    if( reset || !file.exists() ) {
      this.dump("Creating new datasource");
      ds = Components.classes["@mozilla.org/rdf/datasource;1?name=xml-datasource"]
      .createInstance(Components.interfaces.nsIRDFDataSource);

      var rootnode = RDF.GetResource(uri);
      RDFCUtils.MakeSeq(ds, rootnode);
      var con = this.getContainer(ds, rootnode);
      ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
      ds.FlushTo(fileURI);
    } else {
      ds = RDF.GetDataSourceBlocking(fileURI);
    }
    return ds;
  },
  
  readDocument: function(aFile, rootTag)
  {
    //setup input stream    
    var contents = new String();    
    var fiStream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
    var siStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);         
    var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);    
    converter.charset = "UTF-8";    

    try {
      //load file into stram
      fiStream.init(aFile, 1, 0, false);
      siStream.init(fiStream);
      while (siStream.available() > 0) {
	var chunk = siStream.read(siStream.available());
	contents += converter.ConvertToUnicode(chunk);      
      };
      
      //close file
      siStream.close();                          
      fiStream.close();
    } catch (e) {
      return {};
    }

    var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);           
    var doc = parser.parseFromString(contents, "text/xml");

    var connHist = {};
    this.histSize = 0;
    var hosts = doc.getElementsByTagName("host");
    for (var i=0; i<hosts.length; i++) {
      var hostHash = hosts[i].getAttribute("hash");
      if (!(hostHash in connHist)) {
	connHist[hostHash] = new Array();
      }
      var conns = hosts[i].getElementsByTagName("conn");
      var entry = {};
      for (var j=0; j<conns.length; j++) {
	entry.connTime = conns[j].getAttribute("ct");
	entry.elapsedTime = conns[j].getAttribute("et");
	entry.status = conns[j].getAttribute("st");
	entry.status = conns[j].getAttribute("st");
      }
      connHist[hostHash].push(entry);
      this.histSize++;
    }
    return connHist;
  },

  writeConnHist: function()
  {
    this.dump("Writing connHist");
    var file = this.getProfileFile(CONN_HIST_FILE);
    var xml = this.serializeConnHist();

    //setup output stream
    var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);     
    var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);             
    converter.charset = "UTF-8";    
    
    //initialize file stream
    var flags = 0x02 | 0x08 | 0x20;
    foStream.init(file, flags, 0664, 0);  
         
    //use string converter
    var chunk = converter.ConvertFromUnicode(xml);
    foStream.write(chunk, chunk.length);
    var fin = converter.Finish();
    if (fin.length > 0)
      foStream.write(fin, fin.length);
    
    //close file stream
    foStream.close(); 
  },

  start: function()
  {
    // do nothing if already loaded
    if (this.initialized) {
      return;
    }
    
    this.dump("Starting diagAgent");

    // create RDF datasources
    // this.connHist = this.createDatasource(CONN_HIST_FILE, CONN_HIST_URI);
    this.connSummary = this.createDatasource(CONN_SUMMARY_FILE, CONN_SUMMARY_URI);

    var file = this.getProfileFile(CONN_HIST_FILE);
    this.connHist = this.readDocument(file, "connHist");

    // add observers
    var obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    obs.addObserver(this, "http-on-examine-response", true);
    obs.addObserver(this, "diagtool-pref-change", true);
    obs.addObserver(this, "profile-before-change", true);

    // retrieve preferences
    var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
    this.logging = prefs.getBoolPref("extensions.diagtool.logging");

    this.obs = obs;
    this.prefs = prefs;
    this.initialized = true;
    this.startTime = {};
  },
  
  stop: function()
  {
    // do nothing if not loaded
    if (!this.initialized) {
      return;
    }
    
    // clean up held observers etc to avoid leaks.
    try {
      // RDF.UnregisterDataSource(this.connHist);
      this.writeConnHist();
      
      var connSummaryFile = this.getProfileFile(CONN_SUMMARY_FILE);
      this.connSummary = this.connSummary.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
      var fileURI = this.getFileURI(connSummaryFile);
      this.dump("Flushing connection summary to " + fileURI);
      this.connSummary.FlushTo(fileURI);

      this.obs.removeObserver(this, "http-on-examine-response");
      this.obs.removeObserver(this, "diagtool-pref-change");
      this.obs.removeObserver(this, "profile-before-change");
    } catch(e) {};
    
    //destroy variables
    this.initialized = null;
  },

  getFileURI: function (aFile)
  {
    var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    // this.dump("getting URI for file " + aFile);
    var uri = ios.newFileURI(aFile);
    return uri.spec;
  },
  
  getWindow: function ffManager_getWindow()
  {
    //XXX may need to change this code if main window of a supported app is not "navigator:browser"
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
    var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
    var top = wm.getMostRecentWindow("navigator:browser");
    if (!top) {
      switch (this.guid) {
        case "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}":
        case "{3db10fab-e461-4c80-8b97-957ad5f8ea47}":
        default:        
          top = ww.openWindow(null, "chrome://browser/content/browser.xul", "_blank", "chrome,all,dialog=no", "about:blank");
          break;
        case "{86c18b42-e466-45a9-ae7a-9b95ba6f5640}":
          top = ww.openWindow(null, "chrome://navigator/content/navigator.xul", "_blank", "chrome,all,dialog=no", "about:blank");
          break;
      };
    };
    return top;  
  },

  restoreConnHist: function() {
    for (hostHash in this.connHist) {
      for (var i=0; i<this.connHist[hostHash].length; i++) {
	if (!(hostHash in this.oldConnHist)) {
	  this.oldConnHist[hostHash] = new Array();
	}
	this.oldConnHist[hostHash].push(this.connHist[hostHash][i]);
	this.oldHistSize++;
      }
    }

    this.connHist = this.oldConnHist;
    this.histSize = this.oldHistSize;
  },

  receiveResponse: function(aXMLReq, dest, success) {
    if (success == 0) {
      this.restoreConnHist();
      this.dump("Failed to connect to the diagnostic agent.");
      return;
    }

    this.dump("Response: " + aXMLReq.responseText);
    var doc = aXMLReq.responseXML;
    var elems = doc.getElementsByTagName("responseStatus");

    if (elems == null) {
      this.restoreConnHist();
      this.dump("Did not understand response from diagnostic agent.");
      return;
    }

    var status = null;
    var conns = null;
    var failures = null;
    var statusDist = {};
    var lastSuccess = null;
    var lastFailure = null;
    var data = {};

    if (elems.length == 1) {
      status = elems.item(0).lastChild.nodeValue;
      if (status == 1) {
	// diagnosis completed successfully, clear the old history.
	//
	this.oldConnHist = null;
	this.oldHistSize = 0;
      } else {
	this.restoreConnHist();
	this.dump("Diagnostic agent returned status " + status);
	return;
      }
    } else {
      this.restoreConnHist();
      this.dump("elems.length == " + elems.length);
      return;
    }

    this.writeConnHist();
    elems = doc.getElementsByTagName("destInfo");
    if (elems.length == 0) {
      // no data to display
      return;
    }
    if (elems.length == 1) {
      var elem = elems.item(0);
      var d = new Date();
      data.users = parseInt(elem.getAttribute("users"));

      if (elem.hasAttribute("avgLatency")) {
	data.avgLatency = parseInt(elem.getAttribute("avgLatency"));
	data.avgLatency = (data.avgLatency / 1000).toFixed(3) + " s";
      } else {
	data.avgLatency = "---";
      }

      if (elem.hasAttribute("lastSuccess")) {
	lastSuccess = parseInt(elem.getAttribute("lastSuccess"));
	data.lastSuccess = this.dateToString(lastSuccess);
      } else {
	data.lastSuccess = "---";
      }

      if (elem.hasAttribute("lastFailure")) {
	lastFailure = parseInt(elem.getAttribute("lastFailure"));
	data.lastFailure = this.dateToString(lastFailure);
      } else {
	data.lastFailure = "---";
      }
      var statuses = elem.getElementsByTagName("status");
      failures = 0;
      conns = 0;
      for (var i=0; i<statuses.length; i++) {
	var status = parseInt(statuses[i].getAttribute("id"));
	var count  = parseInt(statuses[i].getAttribute("count"));
	conns += count;
	if (status != OK) {
	  failures += count;
	}
	if (!(status in statusDist)) {
	  statusDist[status] = 0;
	}
	statusDist[status] += count;
      }
    }

    data.host = dest;
    if (conns > 0) {
      data.failureRate = (failures / conns * 100).toFixed(2) + "%";
    } else {
      data.failureRate = "---";
    }
    data.conns = conns;
    data.statusDist = statusDist;

    var win = this.getWindow();
//     var browser = win.document.getElementById("content");
    var response = "chrome://diag/content/response.xul";
    // var tab = browser.addTab(response);
    win.openDialog(response, "Server Info - " + dest, "chrome,all,dialog=no", data);
  },

  serializeConnHist: function() {
    var connHist = this.connHist;
    var rootElement = "connHist";

    var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);

    var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].createInstance(Components.interfaces.nsIDOMParser);           
    var doc = parser.parseFromString("<" + rootElement + "/>", "text/xml");
    var rootNode = doc.getElementsByTagName(rootElement)[0];
    for (hostHash in connHist) {
      var hostNode = doc.createElement("host");
      hostNode.setAttribute("hash", hostHash);
      rootNode.appendChild(hostNode);
      this.dump("Hash: " + hostHash + ", conns: " + connHist[hostHash].length);
      for (var i=0; i<connHist[hostHash].length; i++) {
	entry = connHist[hostHash][i];
	var connNode = doc.createElement("conn");
	if ("elapsedTime" in entry) {
	  connNode.setAttribute("et", entry.elapsedTime);
	}
	connNode.setAttribute("st", entry.status);
	connNode.setAttribute("ct", entry.connTime);
	hostNode.appendChild(connNode);
      }
    }

    var xml = serializer.serializeToString(doc);
    return xml;
  },

  md5sum: function(s, base64) {
    // for additional privacy, the user can set the salt.  If you only
    // wish to share diagnoses among your friends, for example, you
    // can all use the same secret salt.
    var privGroup = this.prefs.getIntPref("extensions.diagtool.privacyGroup");
    var salt = "";
    if (privGroup == 1) {
      salt = this.prefs.getCharPref("extensions.diagtool.hashSalt");
    }
    s = salt + DATA_ID + s;

    var stream = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);
    stream.setData(s, s.length);
    const nsICryptoHash = Components.interfaces.nsICryptoHash;
    var hashComp = Components.classes["@mozilla.org/security/hash;1"].createInstance(nsICryptoHash);
    
    hashComp.init(nsICryptoHash.MD5);
    hashComp.updateFromStream(stream, stream.available());
    var hash = hashComp.finish(base64); // encode as base64
    return hash;
    return hashComp;
  },

  createXMLRequest: function(dest, origDest, hashDest) {
    // send the time in case the client's clock is wrong
    var d = new Date();
    var t = d.getTime();
    var text = '<diagRequest'
    text += ' version="' + PROTO_VER + '"'
    text += ' time="' + t + '"'
    text += ' clientVer="' + CLIENT_VER + '"'
    text += '>\n';

    // TODO: let people specify the dest, and give it an id

    if (dest && origDest) {
      text += '<component dest="' + dest + '"/>\n';
    }

    if (dest && hashDest) {
      var hashedDest = this.md5sum(dest, true);
      text += '<component destHash="' + hashedDest + '"/>\n';
    }


    // I'll roll my own XML to reduce the size of the data to transmit
    var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].createInstance(Components.interfaces.nsIDOMSerializer);       

    text += this.serializeConnHist();

    // take a snapshot and clear the conn history
    this.oldConnHist = this.connHist;
    this.oldHistSize = this.histSize;
    this.connHist = {};
    this.histSize = 0;

    text += "</diagRequest>\n";
    return text;
  },

  requestDiagnosis: function(dest, origDest, hashDest) {
    if (dest == null || dest == "") {
      dest = this.currentHost;
    }
    this.sendRequest(dest, origDest, hashDest);
  },
  
  sendRequest: function(dest, origDest, hashDest) {
    // request diagnosis from a diagnostic agent
    // First serialize and send connection history info.
    
    this.dump("Starting diagnostic request to dest " + dest);

    var agent = this;
    
    // don't send the original hostname for now.
    var text = this.createXMLRequest(dest, origDest, hashDest);

    var reqBody = '<?xml version="1.0" encoding="UTF-8" ?>\n'
    reqBody += text;

    var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
    // TODO: need to handle timeouts
    // request.timeout = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    request.overrideMimeType("text/xml");
    request.onerror = function() { agent.receiveResponse(this, dest, 0); };
    request.onload = function() { agent.receiveResponse(this, dest, 1); };

    this.daURI = this.prefs.getCharPref("extensions.diagtool.daURI");
    request.open("POST", this.daURI, true);  // asynchronous
    this.dump("Sending request to " + this.daURI + " with body " + reqBody);

    request.send(reqBody);
  },

  // right now there's no probing.  I need to make this periodic.
  probeHosts: function(hosts, count) {
    var probing = prefs.getBoolPref("extensions.diagtool.baselineProbing");

    if (!probing) {
      return;
    }
    
    function probeCallback(host, status) {
      // create a closure for host and status when the callback is made
      return function() { agent.probeResponse(this, host, status); };
    }

    this.dump("Beginning probes.");
    var agent = this;
    for (var i=0; i<hosts.length; i++) {
      var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);
      // TODO: need to handle timeouts
      // request.timeout = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);

      request.onerror = probeCallback(hosts[i], 0);
      request.onload  = probeCallback(hosts[i], 1);
      request.open("GET", "http://" + hosts[i] + "/status", true);  // asynchronous
      this.dump("Probing " + hosts[i]);
      request.send(null);
    }
  },

  probeResponse: function(aXMLReq, host, success) {
    var d = new Date();
    var curTime = d.getTime();
    this.updateConnHist({host: host, connTime: curTime, status: success});
  },
    
  getConnSummary: function() {
    return this.connSummary;
  },

  getConnHistory: function() {
    // return the history of connections.
    return this.connHist;
  },

  getDateParts: function(t) {
    var d = new Date();
    var curTime = d.setTime(t);

    var rv = {};
    rv.month = d.getMonth() + 1;
    rv.date = d.getDate();
    rv.year = d.getFullYear();
    rv.hour = d.getHours();
    var min = d.getMinutes();
    min = ("0" + min.toString()).substr(-2);
    rv.min = min;
    return rv;
  },
    
  dateToString: function(t) {
//     const dayMillis = 1000 * 60 * 60 * 24;
//     const ageNames = ["Today", "Yesterday"];
//     const daysAgo = "days ago";
    
//     var d = new Date();
//     var now = d.getTime();
//     var curTime = this.getDateParts(now);

    var time = this.getDateParts(t);

//     var today = parseInt(now / dayMillis);
//     var day = parseInt(d / dayMillis);

//     var age = today - day;
    
    
    var s = time.month + "/" + time.date + ", " + time.hour + ":" + time.min;
    return s;
  },

  // update the connection summary based on the new connection
  updateConnSummary: function(entry) {
    var node = RDF.GetResource(PREFIX + "#host-" + entry.host);

    var connsProp          = RDF.GetResource(PREFIX + "#conns");
    var failuresProp       = RDF.GetResource(PREFIX + "#failures");
    var failureRateProp    = RDF.GetResource(PREFIX + "#failureRate");
    var intFailureRateProp = RDF.GetResource(PREFIX + "#intFailureRate");
    var lastConnTimeProp   = RDF.GetResource(PREFIX + "#lastConnTime");
    var lastConnTimeStringProp = RDF.GetResource(PREFIX + "#lastConnTimeString");
    var hostProp         = RDF.GetResource(PREFIX + "#host");

    var conns          = this.connSummary.GetTarget(node, connsProp, true);
    var failures       = this.connSummary.GetTarget(node, failuresProp, true);
    var failureRate    = this.connSummary.GetTarget(node, failureRateProp, true);
    var intFailureRate = this.connSummary.GetTarget(node, intFailureRateProp, true);
    var lastConnTime = this.connSummary.GetTarget(node, lastConnTimeProp, true);
    var lastConnTimeString = this.connSummary.GetTarget(node, lastConnTimeStringProp, true);
    var failure = (entry.status == 0) ? 0 : 1;

    if (conns instanceof Components.interfaces.nsIRDFInt) {
      // update an existing entry
      var c = parseInt(conns.Value) + 1;
      var newConns = RDF.GetIntLiteral(c);
      this.connSummary.Change(node, connsProp, conns, newConns);
      
      if ((failures       instanceof Components.interfaces.nsIRDFInt) &&
	  (failureRate    instanceof Components.interfaces.nsIRDFLiteral) &&
	  (intFailureRate instanceof Components.interfaces.nsIRDFInt) &&
	  (lastConnTime   instanceof Components.interfaces.nsIRDFInt) &&
	  (lastConnTimeString instanceof Components.interfaces.nsIRDFLiteral)) {
	var f = failures.Value + failure;
	var newFailures = RDF.GetIntLiteral(f);
	var rate = f/c*100;
	var intRate = parseInt(f/c*10000);
	var newFailureRate = RDF.GetLiteral(rate.toFixed(2));
	var newIntFailureRate = RDF.GetIntLiteral(intRate);
	var connTime = entry.connTime;
	var connTimeString = this.dateToString(entry.connTime);

	this.connSummary.Change(node, failuresProp, failures, newFailures);
	this.connSummary.Change(node, failureRateProp, failureRate, newFailureRate);
	this.connSummary.Change(node, intFailureRateProp, intFailureRate, newIntFailureRate);
	this.connSummary.Change(node, lastConnTimeProp, lastConnTime,
				RDF.GetIntLiteral(connTime));
	this.connSummary.Change(node, lastConnTimeStringProp, lastConnTimeString,
				RDF.GetLiteral(connTimeString));
	this.obs.notifyObservers(null, "diagtool-update-history", null);
      }
    } else {
      // create a new entry
      var rootnode = RDF.GetResource(CONN_SUMMARY_URI);
      var con = this.getContainer(this.connSummary, rootnode);
      var rate = failure * 100;
      rate = rate.toFixed(2);
      var connTime       = entry.connTime;
      var connTimeString = this.dateToString(entry.connTime);

      this.connSummary.Assert(node, connsProp,
			      RDF.GetIntLiteral(1),            true);
      this.connSummary.Assert(node, hostProp,
			      RDF.GetLiteral(entry.host),      true);
      this.connSummary.Assert(node, lastConnTimeProp,
			      RDF.GetIntLiteral(connTime),     true);
      this.connSummary.Assert(node, lastConnTimeStringProp,
			      RDF.GetLiteral(connTimeString),  true);
      this.connSummary.Assert(node, failuresProp,
			      RDF.GetIntLiteral(failure),      true);
      this.connSummary.Assert(node, failureRateProp,
			      RDF.GetLiteral(rate),            true);
      this.connSummary.Assert(node, intFailureRateProp,
			      RDF.GetIntLiteral(rate*100),     true);
      con.InsertElementAt(node, 1, true);
    }
  },

  resetConnSummary: function() {
    var rootnode = RDF.GetResource(CONN_SUMMARY_URI);
    var con = this.getContainer(this.connSummary, rootnode);
    var elements = con.GetElements();

    while (elements.hasMoreElements()) {
      var element = elements.getNext();
      var preds = this.connSummary.ArcLabelsOut(element);
      while (preds.hasMoreElements()) {
	var pred = preds.getNext();
	var target = this.connSummary.GetTarget(element, pred, true);
	this.connSummary.Unassert(element, pred, target);
      }
      con.RemoveElement(element, false);
    }
  },

  // add a new entry to the connection history
  updateConnHist: function(entry) {
    // this.dump("Updating connHist");
    if (entry.host == 'localhost') {
      this.dump("Not logging localhost");
      return;
    }
    var hostHash = this.md5sum(entry.host, true);
    
    if (!(hostHash in this.connHist)) {
      this.connHist[hostHash] = new Array();
    }

    this.connHist[hostHash].push(entry);
    this.histSize++;

    // this.dump("Done updating connHist");
    if (!("interrupted" in entry)) {
      // don't include interrupted connections in the summary
      this.updateConnSummary(entry);
    }

    // Don't let the histsize grow too much
    if (this.histSize > MAX_HIST_SIZE) {
      this.sendRequest();
    }
  },

  // TODO: make it work with wyciwyg URLs
  extractHost: function(aURI) {
      // regular expression from RFC2396 12.B.
      var reg = new RegExp("^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\\?([^#]*))?(#(.*))?");
      var match = reg.exec(aURI);
      if (match == null) {
	this.dump("Could not parse URI");
	return null;
      }

      var hostname = match[4];
      return hostname;
  },

  /////////////////////////
  // nsIWebProgressListener
  //
  onStateChange: function(aProgress, aRequest, aFlag, aStatus)
  {
    if (!this.initialized) {
      return 0;
    }

    if (!this.logging) {
      return 0;
    }

    // only consider network state changes
    if (!(aFlag & Components.interfaces.nsIWebProgressListener.STATE_IS_NETWORK)) {
      return 0;
    }

    if (aRequest == null) {
      this.dump("Warning: aRequest is null.  aFlag: " + aFlag + ", aStatus: " + aStatus);
      return 0;
    }


    var d = new Date();
    var curTime = d.getTime();

    if (aFlag & STATE_START) {
      // This fires when the load event is initiated
      this.dump("STATE_START");
      aRequest = aRequest.QueryInterface(Components.interfaces.nsIRequest);
      var name = aRequest.name;
      var d = new Date();
      var curTime = d.getTime();
      this.startTime[name] = curTime;
      this.dump("STATE_START URI: " + name);
    }

    if (aFlag & STATE_STOP) {
      // This fires when the load finishes
      aRequest = aRequest.QueryInterface(Components.interfaces.nsIRequest);
      var name = aRequest.name;

      this.dump("STATE_STOP URI: " + name);

      var hostname = this.extractHost(name);
      if (hostname == null) {
	this.dump("Could not parse URI");
	return 0;
      } else {
	this.dump("Host: " + hostname);
      }

      var errorCode = aStatus & 0xffff;
      var status = errorCode;
	
      var entry = {};

      // see mozilla/netwerk/base/public/nsNetError.h for the meaning
      // of the various error codes
     
      switch (errorCode) {
      case 0: 
	// success.  Ignore this because success will be handled by
	// the http-examine-response observer.
      case 3: // Request redirected
      case 4: // retargeted
      case 10: // malformed uri
      case 15: // in progress
      case 16: // offline
      case 17: // no content
      case 18: // unknown protocol
      case 19: // port access not allowed
      case 25: // not resumable
      case 32: // entity changed
      case 31: // redirect loop
	
      case 51: // unknown socket type
      case 52: // socket create failed
	
      case 61: // cache related error codes
      case 62:
      case 63:
      case 64:
      case 65:
      case 66:
      case 67:
      case 68:
      case 70:
	// not network related errors
	this.dump("No network error.  error code: " + errorCode);
	break;
      case 2:
	// Request canceled (e.g. Stop or refresh button pressed).
	// Not necessarily a failure, but if elapsed time is large
	// then it might indicate poor performance.
        entry.interrupted = 1;
      case 13: // connection refused
      case 14: // timeout
	
      case 20: // connection reset
      case 21: // ftp login
      case 22: // ftp cwd
      case 23: // ftp pasv
      case 24: // ftp pwd
      case 25: // ftp list
	
      case 30: // dns lookup failed
	
      case 42: // proxy dns lookup failed
	
      case 71: // net interrupt
      case 72: // proxy connection refused
      
      default:
	entry.host = hostname;
        entry.connTime = curTime;
        entry.status = status;
	
	if (name in this.startTime) {
	  entry.elapsedTime = curTime - this.startTime[name];
	}

        this.dump("**** FAILURE. Host: " + hostname + ", error code: " + errorCode);
	this.currentHost = hostname;
        this.updateConnHist(entry);
      }
    }
    return 0;
  },
  
  onLocationChange: function(aProgress, aRequest, aURI) {
    this.currentHost = aURI.asciiHost;
    this.dump("onLocationChange() URI: " + aURI.spec + ", Host: " + this.currentHost);
  },

  onProgressChange: function() { return 0; },
  onStatusChange: function() { return 0; },
  onSecurityChange: function() {return 0;},
  onLinkIconAvailable: function() {return 0;},

  /////////////////////////////
  // nsIObserver
  observe: function(aSubject, aTopic, aData)
  {
    var enabled, prefix;
    var d = new Date();
    var curTime = d.getTime();
    
    //do nothing until initialized
    if (!this.initialized)
      return;

    if (this.logging && aTopic == "http-on-examine-response") {
      // I don't use EndDocumentLoad here because that also counts
      // cached pages.  http-on-examine response logs all successfully
      // retrieved uncached pages.  I'm going to ignore cached page
      // loads.  status OK is success.
      aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
      var hostname = aSubject.URI.asciiHost;
      var uri = aSubject.URI.spec;
      var entry = {host: hostname,
		   connTime: curTime,
		   status: OK};
      this.dump("http-on-examine-response URI: " + uri);
      if (uri in this.startTime) {
	entry.elapsedTime = curTime - this.startTime[uri];
      }

      this.updateConnHist(entry);
      this.dump("**** SUCCESS: " + hostname);
    }
    else if (aTopic == "diagtool-pref-change") {
      this.logging = (aData == 'true');
    }
    else if (aTopic == "profile-before-change") {
      this.stop();
      this.initialized = false;
    }
  },
     
  ///////////////////////////
  // nsIClassInfo  
  getInterfaces: function(aCount)
  {
    var ifaces = new Array();
    ifaces.push(Components.interfaces.nsIDiagAgent);
    ifaces.push(Components.interfaces.nsIObserver);
    ifaces.push(Components.interfaces.nsIWebProgressListener);
    ifaces.push(Components.interfaces.nsISupportsWeakReference);
    ifaces.push(Components.interfaces.nsIClassInfo);
    ifaces.push(Components.interfaces.nsISupports);
    aCount.value = ifaces.length;
    return ifaces;
  },
  
  getHelperForLanguage: function(aLanguage) { return null; },
  get contractID() { return CONTRACT_ID; },
  get classID() { return CLASS_ID; },
  get classDescription() { return CLASS_NAME; },
  get implementationLanguage() { return Components.interfaces.nsIProgrammingLanguage.JAVASCRIPT; },
  get flags() { return Components.interfaces.nsIClassInfo.SINGLETON; },
   
  ///////////////////////////
  // nsISupports   
  QueryInterface: function(iid)
  {
    if (!iid.equals(Components.interfaces.nsIDiagAgent) &&
        !iid.equals(Components.interfaces.nsIObserver) &&
        !iid.equals(Components.interfaces.nsIClassInfo) &&
        !iid.equals(Components.interfaces.nsIWebProgressListener) &&
        !iid.equals(Components.interfaces.nsISupportsWeakReference) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
}

/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  initialized: false,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (!this.initialized) {
      this.initialized = true;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    };
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);     
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function(aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;

    var agent = new diagAgent();
    agent.start();
    return agent.QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }
