/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ServStats.
 *
 * The Initial Developer of the Original Code is
 * George J. Lee <gjl@csail.mit.edu>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var diagHistory;

function DiagHistory() { this.start(); };

DiagHistory.prototype = {
  ds: null,
  hist: null,
  agent: null,
  tree: null,
  debug: false,

  start: function() {
      this.agent = Components.classes["@capri.csail.mit.edu/ua;1"].getService(Components.interfaces.nsIDiagAgent);
      this.tree = document.getElementById("diag-hist-tree");
      var obs = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
      obs.addObserver(this, "diagtool-update-history", false);
      this.obs = obs;
  },

  stop: function() {
      this.obs.removeObserver(this, "diagtool-update-history");
  },
    
  dump: function(aMessage) {
    if (this.debug) {
      var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
      consoleService.logStringMessage("DiagHistory: " + aMessage);
    }
  },

  requestDiagnosis: function() {
    document.getElementById("diag-hist-button").setAttribute("disabled","true");
    var row = this.tree.currentIndex;
    var col = this.tree.columns.getNamedColumn("diag-hist-dest");
    if (row == -1) {
      this.dump("Warning, no selection made");
      return;
    }
    this.dump("getting dest at row " + row + " and col " + col);
    var dest = this.tree.view.getCellText(row, col);
    
    this.dump("Currently selected host: " + dest);
    this.agent.requestDiagnosis(dest, false, true);
  },

  clearHistory: function() {
    this.dump("clearing history");
    this.agent.resetConnSummary();
    this.updateList();
  },

  updateList: function() {
    if (this.hist == null) {
      this.dump("Starting update");
      // Get the history from the global diagTool object
      this.hist = this.agent.getConnSummary();
      
      this.ds = this.tree.database;
      this.ds.AddDataSource(this.hist);
      this.tree.builder.rebuild();
      this.dump("Rebuilt tree");
    }
  },
  
  updateButton: function() {
    var disabled = (this.tree.currentIndex == -1);
    document.getElementById("diag-hist-button").setAttribute("disabled",disabled);
  },
    
  observe: function(aSubject, aTopic, aData) {
    if (aTopic == "diagtool-update-history" && this.tree != null) {
      this.tree.builder.rebuild();
    }
  }
};

function historyLoad() {
  diagHistory = new DiagHistory();
  diagHistory.updateList();
}

function historyUnload() {
  diagHistory.stop();
  diagHistory = null;
}
